@@ -1,21 +1,35 @@
 """Main Nvim interface."""
+
+from __future__ import annotations
+
+import asyncio
 import os
 import sys
 import threading
 from functools import partial
 from traceback import format_stack
+from types import SimpleNamespace
+from typing import (Any, AnyStr, Callable, Dict, Iterator, List, Optional,
+                    TYPE_CHECKING, Union)
 
 from msgpack import ExtType
 
 from pynvim.api.buffer import Buffer
 from pynvim.api.common import (NvimError, Remote, RemoteApi, RemoteMap, RemoteSequence,
-                               decode_if_bytes, walk)
+                               TDecodeMode, decode_if_bytes, walk)
 from pynvim.api.tabpage import Tabpage
 from pynvim.api.window import Window
-from pynvim.compat import IS_PYTHON3
-from pynvim.util import Version, format_exc_skip
+from pynvim.util import format_exc_skip
 
-__all__ = ('Nvim')
+if TYPE_CHECKING:
+    from pynvim.msgpack_rpc import Session
+
+if sys.version_info < (3, 8):
+    from typing_extensions import Literal
+else:
+    from typing import Literal
+
+__all__ = ['Nvim']
 
 
 os_chdir = os.chdir
@@ -44,8 +58,7 @@ _G["_pynvim_"..chid] = mod
 """
 
 
-class Nvim(object):
-
+class Nvim:
     """Class that represents a remote Nvim instance.
 
     This class is main entry point to Nvim remote API, it is a wrapper
@@ -69,7 +82,7 @@ class Nvim(object):
     """
 
     @classmethod
-    def from_session(cls, session):
+    def from_session(cls, session: Session) -> Nvim:
         """Create a new Nvim instance for a Session instance.
 
         This method must be called to create the first Nvim instance, since it
@@ -79,9 +92,7 @@ class Nvim(object):
         session.error_wrapper = lambda e: NvimError(decode_if_bytes(e[1]))
         channel_id, metadata = session.request(b'nvim_get_api_info')
 
-        if IS_PYTHON3:
-            # decode all metadata strings for python3
-            metadata = walk(decode_if_bytes, metadata)
+        metadata = walk(decode_if_bytes, metadata)
 
         types = {
             metadata['types']['Buffer']['id']: Buffer,
@@ -92,40 +103,58 @@ class Nvim(object):
         return cls(session, channel_id, metadata, types)
 
     @classmethod
-    def from_nvim(cls, nvim):
+    def from_nvim(cls, nvim: Nvim) -> Nvim:
         """Create a new Nvim instance from an existing instance."""
         return cls(nvim._session, nvim.channel_id, nvim.metadata,
                    nvim.types, nvim._decode, nvim._err_cb)
 
-    def __init__(self, session, channel_id, metadata, types,
-                 decode=False, err_cb=None):
+    def __init__(
+        self,
+        session: Session,
+        channel_id: int,
+        metadata: Dict[str, Any],
+        types: Dict[int, Any],
+        decode: TDecodeMode = True,
+        err_cb: Optional[Callable[[str], None]] = None
+    ):
         """Initialize a new Nvim instance. This method is module-private."""
         self._session = session
         self.channel_id = channel_id
         self.metadata = metadata
         version = metadata.get("version", {"api_level": 0})
-        self.version = Version(**version)
+        self.version = SimpleNamespace(**version)
         self.types = types
         self.api = RemoteApi(self, 'nvim_')
         self.vars = RemoteMap(self, 'nvim_get_var', 'nvim_set_var', 'nvim_del_var')
         self.vvars = RemoteMap(self, 'nvim_get_vvar', None, None)
         self.options = RemoteMap(self, 'nvim_get_option', 'nvim_set_option')
         self.buffers = Buffers(self)
-        self.windows = RemoteSequence(self, 'nvim_list_wins')
-        self.tabpages = RemoteSequence(self, 'nvim_list_tabpages')
+        self.windows: RemoteSequence[Window] = RemoteSequence(self, 'nvim_list_wins')
+        self.tabpages: RemoteSequence[Tabpage] = RemoteSequence(
+            self, 'nvim_list_tabpages'
+        )
         self.current = Current(self)
         self.session = CompatibilitySession(self)
         self.funcs = Funcs(self)
         self.lua = LuaFuncs(self)
         self.error = NvimError
         self._decode = decode
-        self._err_cb = err_cb
+        if err_cb is None:
+            self._err_cb: Callable[[str], Any] = lambda _: None
+        else:
+            self._err_cb = err_cb
 
-        # only on python3.4+ we expose asyncio
-        if IS_PYTHON3:
-            self.loop = self._session.loop._loop
+    @property
+    def loop(self) -> asyncio.AbstractEventLoop:
+        """Get the event loop (exposed to rplugins)."""  # noqa
 
-    def _from_nvim(self, obj, decode=None):
+        # see #294: for python 3.4+, the only available and guaranteed
+        # implementation of msgpack_rpc BaseEventLoop is the AsyncioEventLoop.
+        # The underlying asyncio event loop is exposed to rplugins.
+        # pylint: disable=protected-access
+        return self._session.loop._loop  # type: ignore
+
+    def _from_nvim(self, obj: Any, decode: Optional[TDecodeMode] = None) -> Any:
         if decode is None:
             decode = self._decode
         if type(obj) is ExtType:
@@ -135,18 +164,18 @@ class Nvim(object):
             obj = decode_if_bytes(obj, decode)
         return obj
 
-    def _to_nvim(self, obj):
+    def _to_nvim(self, obj: Any) -> Any:
         if isinstance(obj, Remote):
             return ExtType(*obj.code_data)
         return obj
 
-    def _get_lua_private(self):
+    def _get_lua_private(self) -> LuaFuncs:
         if not getattr(self._session, "_has_lua", False):
             self.exec_lua(lua_module, self.channel_id)
-            self._session._has_lua = True
+            self._session._has_lua = True  # type: ignore[attr-defined]
         return getattr(self.lua, "_pynvim_{}".format(self.channel_id))
 
-    def request(self, name, *args, **kwargs):
+    def request(self, name: str, *args: Any, **kwargs: Any) -> Any:
         r"""Send an API request or notification to nvim.
 
         It is rarely needed to call this function directly, as most API
@@ -180,9 +209,9 @@ class Nvim(object):
         decode = kwargs.pop('decode', self._decode)
         args = walk(self._to_nvim, args)
         res = self._session.request(name, *args, **kwargs)
-        return walk(self._from_nvim, res, decode=decode)
+        return walk(partial(self._from_nvim, decode=decode), res)
 
-    def next_message(self):
+    def next_message(self) -> Any:
         """Block until a message(request or notification) is available.
 
         If any messages were previously enqueued, return the first in queue.
@@ -192,8 +221,13 @@ class Nvim(object):
         if msg:
             return walk(self._from_nvim, msg)
 
-    def run_loop(self, request_cb, notification_cb,
-                 setup_cb=None, err_cb=None):
+    def run_loop(
+        self,
+        request_cb: Optional[Callable[[str, List[Any]], Any]],
+        notification_cb: Optional[Callable[[str, List[Any]], Any]],
+        setup_cb: Optional[Callable[[], None]] = None,
+        err_cb: Optional[Callable[[str], Any]] = None
+    ) -> None:
         """Run the event loop to receive requests and notifications from Nvim.
 
         This should not be called from a plugin running in the host, which
@@ -203,11 +237,11 @@ class Nvim(object):
             err_cb = sys.stderr.write
         self._err_cb = err_cb
 
-        def filter_request_cb(name, args):
+        def filter_request_cb(name: str, args: Any) -> Any:
             name = self._from_nvim(name)
             args = walk(self._from_nvim, args)
             try:
-                result = request_cb(name, args)
+                result = request_cb(name, args)  # type: ignore[misc]
             except Exception:
                 msg = ("error caught in request handler '{} {}'\n{}\n\n"
                        .format(name, args, format_exc_skip(1)))
@@ -215,11 +249,11 @@ class Nvim(object):
                 raise
             return walk(self._to_nvim, result)
 
-        def filter_notification_cb(name, args):
+        def filter_notification_cb(name: str, args: Any) -> None:
             name = self._from_nvim(name)
             args = walk(self._from_nvim, args)
             try:
-                notification_cb(name, args)
+                notification_cb(name, args)  # type: ignore[misc]
             except Exception:
                 msg = ("error caught in notification handler '{} {}'\n{}\n\n"
                        .format(name, args, format_exc_skip(1)))
@@ -228,31 +262,33 @@ class Nvim(object):
 
         self._session.run(filter_request_cb, filter_notification_cb, setup_cb)
 
-    def stop_loop(self):
+    def stop_loop(self) -> None:
         """Stop the event loop being started with `run_loop`."""
         self._session.stop()
 
-    def close(self):
+    def close(self) -> None:
         """Close the nvim session and release its resources."""
         self._session.close()
 
-    def __enter__(self):
+    def __enter__(self) -> Nvim:
         """Enter nvim session as a context manager."""
         return self
 
-    def __exit__(self, *exc_info):
+    def __exit__(self, *exc_info: Any) -> None:
         """Exit nvim session as a context manager.
 
         Closes the event loop.
         """
         self.close()
 
-    def with_decode(self, decode=True):
+    def with_decode(self, decode: Literal[True] = True) -> Nvim:
         """Initialize a new Nvim instance."""
         return Nvim(self._session, self.channel_id,
                     self.metadata, self.types, decode, self._err_cb)
 
-    def ui_attach(self, width, height, rgb=None, **kwargs):
+    def ui_attach(
+        self, width: int, height: int, rgb: Optional[bool] = None, **kwargs: Any
+    ) -> None:
         """Register as a remote UI.
 
         After this method is called, the client will receive redraw
@@ -263,42 +299,42 @@ class Nvim(object):
             options['rgb'] = rgb
         return self.request('nvim_ui_attach', width, height, options)
 
-    def ui_detach(self):
+    def ui_detach(self) -> None:
         """Unregister as a remote UI."""
         return self.request('nvim_ui_detach')
 
-    def ui_try_resize(self, width, height):
+    def ui_try_resize(self, width: int, height: int) -> None:
         """Notify nvim that the client window has resized.
 
         If possible, nvim will send a redraw request to resize.
         """
         return self.request('ui_try_resize', width, height)
 
-    def subscribe(self, event):
+    def subscribe(self, event: str) -> None:
         """Subscribe to a Nvim event."""
         return self.request('nvim_subscribe', event)
 
-    def unsubscribe(self, event):
+    def unsubscribe(self, event: str) -> None:
         """Unsubscribe to a Nvim event."""
         return self.request('nvim_unsubscribe', event)
 
-    def command(self, string, **kwargs):
+    def command(self, string: str, **kwargs: Any) -> None:
         """Execute a single ex command."""
         return self.request('nvim_command', string, **kwargs)
 
-    def command_output(self, string):
+    def command_output(self, string: str) -> str:
         """Execute a single ex command and return the output."""
         return self.request('nvim_command_output', string)
 
-    def eval(self, string, **kwargs):
+    def eval(self, string: str, **kwargs: Any) -> Any:
         """Evaluate a vimscript expression."""
         return self.request('nvim_eval', string, **kwargs)
 
-    def call(self, name, *args, **kwargs):
+    def call(self, name: str, *args: Any, **kwargs: Any) -> Any:
         """Call a vimscript function."""
         return self.request('nvim_call_function', name, args, **kwargs)
 
-    def exec_lua(self, code, *args, **kwargs):
+    def exec_lua(self, code: str, *args: Any, **kwargs: Any) -> Any:
         """Execute lua code.
 
         Additional parameters are available as `...` inside the lua chunk.
@@ -319,18 +355,18 @@ class Nvim(object):
         """
         return self.request('nvim_execute_lua', code, args, **kwargs)
 
-    def strwidth(self, string):
+    def strwidth(self, string: str) -> int:
         """Return the number of display cells `string` occupies.
 
         Tab is counted as one cell.
         """
         return self.request('nvim_strwidth', string)
 
-    def list_runtime_paths(self):
+    def list_runtime_paths(self) -> List[str]:
         """Return a list of paths contained in the 'runtimepath' option."""
         return self.request('nvim_list_runtime_paths')
 
-    def foreach_rtp(self, cb):
+    def foreach_rtp(self, cb: Callable[[str], Any]) -> None:
         """Invoke `cb` for each path in 'runtimepath'.
 
         Call the given callable for each path in 'runtimepath' until either
@@ -338,30 +374,23 @@ class Nvim(object):
         are no longer paths. If stopped in case callable returned non-None,
         vim.foreach_rtp function returns the value returned by callable.
         """
-        for path in self.request('nvim_list_runtime_paths'):
+        for path in self.list_runtime_paths():
             try:
                 if cb(path) is not None:
                     break
             except Exception:
                 break
 
-    def chdir(self, dir_path):
+    def chdir(self, dir_path: str) -> None:
         """Run os.chdir, then all appropriate vim stuff."""
         os_chdir(dir_path)
         return self.request('nvim_set_current_dir', dir_path)
 
-    def feedkeys(self, keys, options='', escape_csi=True):
-        """Push `keys` to Nvim user input buffer.
-
-        Options can be a string with the following character flags:
-        - 'm': Remap keys. This is default.
-        - 'n': Do not remap keys.
-        - 't': Handle keys as if typed; otherwise they are handled as if coming
-               from a mapping. This matters for undo, opening folds, etc.
-        """
+    def feedkeys(self, keys: str, options: str = '', escape_csi: bool = True) -> None:
+        """Push `keys` to Nvim user input buffer."""
         return self.request('nvim_feedkeys', keys, options, escape_csi)
 
-    def input(self, bytes):
+    def input(self, bytes: AnyStr) -> int:
         """Push `bytes` to Nvim low level input buffer.
 
         Unlike `feedkeys()`, this uses the lowest level input buffer and the
@@ -371,8 +400,13 @@ class Nvim(object):
         """
         return self.request('nvim_input', bytes)
 
-    def replace_termcodes(self, string, from_part=False, do_lt=True,
-                          special=True):
+    def replace_termcodes(
+        self,
+        string: str,
+        from_part: bool = False,
+        do_lt: bool = True,
+        special: bool = True
+    ) -> str:
         r"""Replace any terminal code strings by byte sequences.
 
         The returned sequences are Nvim's internal representation of keys,
@@ -388,14 +422,14 @@ class Nvim(object):
         return self.request('nvim_replace_termcodes', string,
                             from_part, do_lt, special)
 
-    def out_write(self, msg, **kwargs):
+    def out_write(self, msg: str, **kwargs: Any) -> None:
         r"""Print `msg` as a normal message.
 
         The message is buffered (won't display) until linefeed ("\n").
         """
         return self.request('nvim_out_write', msg, **kwargs)
 
-    def err_write(self, msg, **kwargs):
+    def err_write(self, msg: str, **kwargs: Any) -> None:
         r"""Print `msg` as an error message.
 
         The message is buffered (won't display) until linefeed ("\n").
@@ -408,11 +442,11 @@ class Nvim(object):
             return
         return self.request('nvim_err_write', msg, **kwargs)
 
-    def _thread_invalid(self):
+    def _thread_invalid(self) -> bool:
         return (self._session._loop_thread is not None
                 and threading.current_thread() != self._session._loop_thread)
 
-    def quit(self, quit_command='qa!'):
+    def quit(self, quit_command: str = 'qa!') -> None:
         """Send a quit command to Nvim.
 
         By default, the quit command is 'qa!' which will make Nvim quit without
@@ -426,11 +460,11 @@ class Nvim(object):
             # ignore it.
             pass
 
-    def new_highlight_source(self):
+    def new_highlight_source(self) -> int:
         """Return new src_id for use with Buffer.add_highlight."""
         return self.current.buffer.add_highlight("", 0, src_id=0)
 
-    def async_call(self, fn, *args, **kwargs):
+    def async_call(self, fn: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
         """Schedule `fn` to be called by the event loop soon.
 
         This function is thread-safe, and is the only way code not
@@ -442,7 +476,7 @@ class Nvim(object):
         """
         call_point = ''.join(format_stack(None, 5)[:-1])
 
-        def handler():
+        def handler() -> None:
             try:
                 fn(*args, **kwargs)
             except Exception as err:
@@ -465,26 +499,26 @@ class Buffers(object):
     Conforms to *python-buffers*.
     """
 
-    def __init__(self, nvim):
+    def __init__(self, nvim: Nvim):
         """Initialize a Buffers object with Nvim object `nvim`."""
         self._fetch_buffers = nvim.api.list_bufs
 
-    def __len__(self):
+    def __len__(self) -> int:
         """Return the count of buffers."""
         return len(self._fetch_buffers())
 
-    def __getitem__(self, number):
+    def __getitem__(self, number: int) -> Buffer:
         """Return the Buffer object matching buffer number `number`."""
         for b in self._fetch_buffers():
             if b.number == number:
                 return b
         raise KeyError(number)
 
-    def __contains__(self, b):
+    def __contains__(self, b: Buffer) -> bool:
         """Return whether Buffer `b` is a known valid buffer."""
         return isinstance(b, Buffer) and b.valid
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[Buffer]:
         """Return an iterator over the list of buffers."""
         return iter(self._fetch_buffers())
 
@@ -493,7 +527,7 @@ class CompatibilitySession(object):
 
     """Helper class for API compatibility."""
 
-    def __init__(self, nvim):
+    def __init__(self, nvim: Nvim):
         self.threadsafe_call = nvim.async_call
 
 
@@ -501,72 +535,70 @@ class Current(object):
 
     """Helper class for emulating vim.current from python-vim."""
 
-    def __init__(self, session):
+    def __init__(self, session: Nvim):
         self._session = session
         self.range = None
 
     @property
-    def line(self):
+    def line(self) -> str:
         return self._session.request('nvim_get_current_line')
 
     @line.setter
-    def line(self, line):
+    def line(self, line: str) -> None:
         return self._session.request('nvim_set_current_line', line)
 
     @line.deleter
-    def line(self):
+    def line(self) -> None:
         return self._session.request('nvim_del_current_line')
 
     @property
-    def buffer(self):
+    def buffer(self) -> Buffer:
         return self._session.request('nvim_get_current_buf')
 
     @buffer.setter
-    def buffer(self, buffer):
+    def buffer(self, buffer: Union[Buffer, int]) -> None:
         return self._session.request('nvim_set_current_buf', buffer)
 
     @property
-    def window(self):
+    def window(self) -> Window:
         return self._session.request('nvim_get_current_win')
 
     @window.setter
-    def window(self, window):
+    def window(self, window: Union[Window, int]) -> None:
         return self._session.request('nvim_set_current_win', window)
 
     @property
-    def tabpage(self):
+    def tabpage(self) -> Tabpage:
         return self._session.request('nvim_get_current_tabpage')
 
     @tabpage.setter
-    def tabpage(self, tabpage):
+    def tabpage(self, tabpage: Union[Tabpage, int]) -> None:
         return self._session.request('nvim_set_current_tabpage', tabpage)
 
 
-class Funcs(object):
-
+class Funcs:
     """Helper class for functional vimscript interface."""
 
-    def __init__(self, nvim):
+    def __init__(self, nvim: Nvim):
         self._nvim = nvim
 
-    def __getattr__(self, name):
+    def __getattr__(self, name: str) -> Callable[..., Any]:
         return partial(self._nvim.call, name)
 
 
-class LuaFuncs(object):
-
+class LuaFuncs:
     """Wrapper to allow lua functions to be called like python methods."""
 
-    def __init__(self, nvim, name=""):
+    def __init__(self, nvim: Nvim, name: str = ""):
         self._nvim = nvim
         self.name = name
 
-    def __getattr__(self, name):
+    def __getattr__(self, name: str) -> LuaFuncs:
         """Return wrapper to named api method."""
         prefix = self.name + "." if self.name else ""
         return LuaFuncs(self._nvim, prefix + name)
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: Any, **kwargs: Any) -> Any:
         # first new function after keyword rename, be a bit noisy
         if 'async' in kwargs:
             raise ValueError('"async" argument is not allowed. '
